import signal

class UserTimeoutError(Exception):
    def __str__(self):
        return "Timed Out"

def timeout(t):
    def decorate(f):
        def handler(signum, frame):
            raise UserTimeoutError()
        def new_f(*args, **kwargs):
            old = signal.signal(signal.SIGALRM, handler)
            assert old is signal.SIG_DFL
            signal.setitimer(signal.ITIMER_REAL, t)
            try:
                result = f(*args, **kwargs)
            finally:
                try:
                    signal.setitimer(signal.ITIMER_REAL, 0)
                    signal.signal(signal.SIGALRM, old)
                except UserTimeoutError:
                    signal.signal(signal.SIGALRM, old)
            return result
        return new_f
    return decorate

import time



if __name__ == '__main__':

    @timeout(3.3)
    def mytest():
        print("Start")
        for i in range(1, 10):
            time.sleep(1)
            print("%d seconds have passed" % i)

    mytest()


from blackopt.abc.solver import Solver, Problem, Solution
import math
import random


class SimAnneal(Solver):
    name = "Simulated Annealing"
    def __init__(self, problem: Problem, solution_cls, mutation_rate, temp = 0.5, half_times=20):

        super().__init__(problem, solution_cls)

        self.mutation_rate = mutation_rate
        self.temp_initial = temp
        self.temp = None
        self.half_times = half_times
        self.alpha = None
        self.current_solution = self.best_solution


    @property
    def cur_fitness(self):
        return self.current_solution.score

    @property
    def best_fitness(self):
        return self.best_solution.score


    def p_accept(self, candidate_fitness):
        """
        Probability of accepting if the candidate is worse than current
        Depends on the current temperature and difference between candidate and current
        """
        return math.exp(-abs(candidate_fitness - self.cur_fitness) / (self.problem.score_span * self.temp) )

    def maybe_accept(self, candidate: Solution):
        """
        Accept with probability 1 if candidate is better than current
        Accept with probabilty p_accept(..) if candidate is worse
        """
        candidate_fitness = candidate.score
        if candidate_fitness > self.cur_fitness:
            self.current_solution = candidate
            if candidate_fitness > self.best_fitness:
                self.best_solution = candidate

        else:
            if random.random() < self.p_accept(candidate_fitness):
                self.current_solution = candidate
            else:
                self.current_solution = self.best_solution

    def solve(self, steps):
        """
        Execute simulated annealing algorithm
        """
        self.problem.eval_count = 0
        self.temp = self.temp_initial
        self.alpha = 2 ** (- self.half_times / steps)
        doc_freq = 1 + steps // 500

        while self.problem.eval_count < steps:
            candidate = self.current_solution.mutate(self.mutation_rate)
            self.maybe_accept(candidate)
            self.temp *= self.alpha
            if not self.problem.eval_count % doc_freq :
                self.record()

    def __str__(self):
        return f"{self.name} mr-{self.mutation_rate} ht-{self.half_times} t-{self.temp_initial}"





from typing import List
import random
from functools import lru_cache

from blackopt.abc import Problem, Solution

import math
from collections import namedtuple

sin_expr = namedtuple("sin_expr", "i j k")


def evaluate_sin(solution: List[float], expr: sin_expr) -> float:
    arg = solution[expr.i] * solution[expr.j] + expr.k
    return math.sin(arg)


class BumpyProblem(Problem):
    """
    fitness = Sum ( sin(x_i * x_j + k) )
    """

    def __init__(self, n_dim: int, expressions: List[sin_expr]):

        self.n_dim = n_dim
        self.expressions = expressions
        self.score_span = 2 * len(expressions)
        self.eval_count = 0

    @staticmethod
    def random_problem(n_dim: int, n_expr: int):

        exprs = [
            sin_expr(
                random.randint(0, n_dim - 1),
                random.randint(0, n_dim - 1),
                random.random(),
            )
            for _ in range(n_expr)
        ]

        return BumpyProblem(n_dim, exprs)

    def evaluate(self, s: 'BumpySolution') -> int:
        self.eval_count += 1
        return sum(evaluate_sin(s.genes, expr) for expr in self.expressions)

    def __str__(self):
        return f"{self.__class__.__name__} {self.n_dim} {len(self.expressions)}"


class BumpySolution(Solution):
    def __init__(self, values):
        self.genes = values

    @staticmethod
    def random_solution() -> Solution:
        values = [random.random() for i in range(BumpySolution.problem.n_dim)]
        return BumpySolution(values)

    def mutate(self, rate: float):
        new_values = []
        for v in self.genes:
            if random.random() < rate:
                new_values.append(random.random())
            else:
                new_values.append(v)

        return BumpySolution(new_values)

    def crossover(self, other: 'BumpySolution'):
        crossover_point = random.randint(1, len(self.genes) - 1)

        child_a = self.genes[:crossover_point] + other.genes[crossover_point:]
        child_b = other.genes[:crossover_point] + self.genes[crossover_point:]

        return [BumpySolution(child_a), BumpySolution(child_b)]

    @lru_cache(maxsize=512)
    def similarity(self, other: 'BumpySolution'):

        diff = 0
        for i in range(self.problem.n_dim):
            diff += abs(self.genes[i] - other.genes[i])

        return 1 - diff / self.problem.n_dim



from collections import defaultdict
from typing import List

import numpy as np

from blackopt.abc import Problem, Solution
from blackopt.abc.solver import Solver



class GeneticAlgorithm(Solver):
    name = "GA"

    def __init__(
        self,
        problem: Problem,
        solution_cls,
        popsize: int,
        mutation_rate: float,
        elite_size: int = 0,
        equal_chances: float = 0.5
    ):

        assert 0 < mutation_rate <= 1
        assert 0 <= equal_chances <= 1
        assert popsize > 1
        assert popsize > elite_size
        assert isinstance(elite_size, int)

        self.mutation_rate = mutation_rate
        self.popsize = popsize
        self.elite_size = elite_size
        self.equal_chances = equal_chances

        super().__init__(problem, solution_cls)

        self.population: List[Solution] = [solution_cls.random_solution() for _ in range(popsize)]
        self.generation = 1
        self.avg = None
        self._rank()

    @property
    def actual_popsize(self):
        return len(self.population)

    def solve(self, steps):
        self.problem.eval_count = 0
        while self.problem.eval_count < steps:

            next_generation = self.population[: self.elite_size]
            next_generation += self._breed(self.popsize - self.elite_size)
            self.population = next_generation

            self._rank()
            self.record()
            self.generation += 1
            if not self.generation % 100:
                print("Generation", self.generation, self.problem.eval_count)

        self.salut()

    def salut(self):
        print(f"{self} is Done in {self.generation} generations / {self.problem.eval_count} evaluations.")

    def _rank(self):
        self.population = sorted(self.population, key=lambda x: x.score + x.regularization_score, reverse=True)
        self.best_solution: Solution = max(self.population + [self.best_solution], key=lambda x: x.score)

    def record(self):
        super().record()
        aggr = defaultdict(list)
        for p in self.population:
            ms = p.metrics()
            for k, v in ms.items():
                aggr[k].append(v)
        for k, lst in aggr.items():
            self.record_metric(
                f"average_{k}",
                sum(lst) / self.actual_popsize,
            )

    def _select_parents(self, n: int) -> List[Solution]:
        indexes = np.arange(0, len(self.population), dtype=np.int)
        chances = np.arange(
            len(self.population), 0, -1, dtype=np.int
        ) + self.equal_chances / (1 - self.equal_chances + 1e-9)
        chances = chances / sum(chances)
        parent_indexes = np.random.choice(indexes, n, True, chances)
        parents = np.array(self.population)[parent_indexes]

        return parents

    def _breed(self, n: int) -> List[Solution]:

        parents = self._select_parents(n)
        children: List[Solution] = []

        for i in range(n):
            parent_1 = parents[i]
            parent_2 = parents[len(parents) - i - 1]
            children += parent_1.crossover(parent_2)

        children = [child.mutate(self.mutation_rate) for child in children]

        return children

    def __str__(self):
        return (
            f"{self.name} with mut_rate - {self.mutation_rate:.5f} & "
            f"pop_size - {self.popsize} & "
            f"elite - {self.elite_size} & equal_c - {self.equal_chances}"
        )



from typing import List, TYPE_CHECKING, ClassVar, Dict, DefaultDict
from collections import defaultdict

import pathos

if TYPE_CHECKING:
    from blackopt.abc import Solver
    from ilya_ezplot import Metric


class SolverFactory:
    def __init__(self, target_cls: ClassVar['Solver'], *args, **kwargs):
        self.target_cls = target_cls
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.target_cls(*self.args, **self.kwargs)


def one_trial(steps: int, solver_constructor: SolverFactory) -> DefaultDict[str, 'Metric']:
    s: Solver = solver_constructor()
    print(s)
    s.solve(steps)
    return s.metrics


def n_runs(trials: int, steps: int, solver: SolverFactory) -> 'Metric':

    pool = pathos.pools.ProcessPool()
    metrics = pool.map(lambda x: one_trial(steps, solver), "x" * trials)

    return sum(metrics)


def compare_solvers(
    trials: int, steps: int, solvers: List[SolverFactory]
) -> Dict[SolverFactory, Dict[str, 'Metric']]:
    pool = pathos.pools.ProcessPool()

    to_map = solvers * trials
    metrics: List[Dict[str, 'Metric']] = pool.map(
        lambda solver: one_trial(steps, solver), solvers * trials
    )
    solver_to_metrics = defaultdict(lambda :defaultdict(list))
    for sf, ms in zip(to_map, metrics):
        for key, metric in ms.items():
            solver_to_metrics[sf][key].append(metric)

    result = defaultdict(dict)
    for sf, metrics_dict in solver_to_metrics.items():
        for key, lst in metrics_dict.items():
            result[sf][key] = sum(lst)

    return result





from __future__ import annotations
from blackopt.algorithms import HillClimber
from blackopt.examples.problems import BumpyProblem, BumpySolution
from blackopt.algorithms import RandomSearch
from blackopt.compare import compare_solvers, SolverFactory
from blackopt.util.document import generate_report


problem = BumpyProblem.random_problem(100, 200)


solvers = [
    SolverFactory(HillClimber, problem, BumpySolution, mutation_rate=mr) for mr in [0.01, 0.03, 0.1]
]
solvers.append(SolverFactory(RandomSearch, problem, BumpySolution))

n_steps = int(1e4)
n_trials = 12

import time
t = time.time()
ms = compare_solvers(n_trials, n_steps, solvers)
print(time.time() - t)

generate_report(problem, ms)





import abc
from typing import TYPE_CHECKING, List, Dict, SupportsFloat

if TYPE_CHECKING:
    from blackopt.abc import Problem


class Solution(abc.ABC):
    problem: 'Problem' = None
    _score: float = None
    regularization_score = 0

    @staticmethod
    @abc.abstractmethod
    def random_solution() -> 'Solution':
        raise NotImplementedError()

    @abc.abstractmethod
    def mutate(self, rate: float) -> 'Solution':
        raise NotImplementedError()

    @abc.abstractmethod
    def crossover(self, other: 'Solution') -> List['Solution']:
        raise NotImplementedError()

    @property
    def score(self):
        if self._score is None:
            self._score = self.problem.evaluate(self)
        return self._score

    def similarity(self, other: 'Solution'):
        if other is self or other == self:
            return 1
        else:
            return 0

    def metrics(self) -> Dict[str, SupportsFloat]:
        return {"score": self.score}



from blackopt.algorithms import HillClimber, GeneticAlgorithm
from blackopt.examples.problems import StepProblem, StepSolution
from blackopt.algorithms import RandomSearch
from blackopt.util.document import generate_report

from blackopt.compare import compare_solvers, SolverFactory


problems = [StepProblem.random_problem(n) for n in [250, 1000, 3000]]

sfs = []

n_steps = 500
trials = 10

for problem in problems:
    sfs += [SolverFactory(GeneticAlgorithm, problem, StepSolution, p, 1/problem.n_dim, e) for p in [10, 30] for e in [0, 1]]
    sfs += [SolverFactory(HillClimber,problem, StepSolution, mutation_rate=2/problem.n_dim)]
    # sfs.append(SolverFactory(RandomSearch, problem, StepSolution))

    ms = compare_solvers(trials, n_steps, sfs)
    generate_report(problem, ms)



from typing import List, Tuple

from blackopt.abc.solver import Solver
from blackopt.algorithms import GeneticAlgorithm

import pathos

import random


def iteration(inp: Tuple[GeneticAlgorithm, int]):
    ga, steps = inp
    ga.solution_cls.problem = ga.problem
    ga.record = lambda : None
    ga.problem.eval_count = 0
    ga.solve(steps)
    ga.metrics = None
    return ga

class MulticoreGeneticAlgorithm(Solver):
    name = "Multicore_GA"

    def __init__(self, problem, solution_cls, *args, **kwargs):

        super().__init__(problem, solution_cls)
        self.args = args
        self.kwargs = kwargs
        self.pool = pathos.pools.ProcessPool()
        self.gas: List[GeneticAlgorithm] = [GeneticAlgorithm(problem, solution_cls, *args, **kwargs) for i in range(self.pool.ncpus)]

    def solve(self, steps):
        self.problem.eval_count = 0
        mapping_steps = int(steps ** (1 / 5))
        steps_per_pool = steps // (self.pool.ncpus * mapping_steps)

        print(steps_per_pool)
        print(mapping_steps)


        for i in range(mapping_steps):
            self.gas = self.pool.map(iteration, zip(self.gas, [steps_per_pool] * len(self.gas) ))

            if i > mapping_steps // 2:
                all_population = sum( (ga.population for ga in self.gas), [])
                random.shuffle(all_population)
                for i, ga in enumerate(self.gas):
                    ga.population = all_population[ga.popsize * i: ga.popsize*(i+1)]
                    ga._rank()

            self.best_solution = max(
                [s.best_solution for s in self.gas] + [self.best_solution], key=lambda x: x.score
            )
            self.problem.eval_count += steps_per_pool * self.pool.ncpus
            self.record()
            print(i * steps_per_pool * self.pool.ncpus)









from typing import List

from blackopt.abc import Solution, Problem
from blackopt.algorithms import GeneticAlgorithm


def keep(child_score: float, parent_min: float, diff: float, pressure: float):
    return child_score > parent_min + pressure * diff


class Gaos(GeneticAlgorithm):
    name = "Gaos"

    def __init__(
        self,
        problem: Problem,
        solution_cls,
        popsize: int,
        mutation_rate: float,
        elite_size: int = 0,
        equal_chances: float = 0.5,
        max_selective_pressure = 200,
        early_stop = True,
    ):
        super().__init__(problem, solution_cls, popsize, mutation_rate, elite_size, equal_chances)
        self.max_selective_pressure = max_selective_pressure
        self.early_stop = early_stop

    def check_early_stop(self):
        return self.selective_pressure >= self.max_selective_pressure

    def solve(self, steps):
        self.problem.eval_count = 0
        while self.problem.eval_count < steps and self.actual_popsize:

            next_generation = self.population[: self.elite_size]
            next_generation += self._breed(
                self.popsize - self.elite_size,
                pressure=0.8 * self.problem.eval_count / steps,
            )
            self.population = next_generation

            self._rank()
            self.record()
            self.generation += 1
            if not self.generation % 10:
                print("Generation", self.generation, self.problem.eval_count)

            if self.early_stop and self.check_early_stop():
                break

        self.salut()


    def record(self):
        super().record()
        self.record_metric("selective pressure", self.selective_pressure)


    def _breed(self, n: int, pressure=0.5) -> List[Solution]:

        result: List[Solution] = []
        ctr = 0
        while len(result) < n and ctr < self.max_selective_pressure:
            ctr += 1
            children = []
            parents = self._select_parents(n)

            parent_scores = {}  # child -> min_parent_score, parents_diff
            for i in range(n):
                parent_1 = parents[i]
                parent_2 = parents[len(parents) - i - 1]
                new = parent_1.crossover(parent_2)
                mutated = []
                for c in new:
                    c = c.mutate(self.mutation_rate)
                    mutated.append(c)
                    parent_scores[c] = (
                        min([parent_1.score, parent_2.score]),
                        abs(parent_1.score - parent_2.score),
                    )

                children += mutated

            result += [
                c for c in children if keep(c.score, *parent_scores[c], pressure)
            ]

        self.selective_pressure = ctr

        return result[:n]



from blackopt.algorithms import Gaos
from blackopt.abc import Problem
import random


def average_similarity(new, pop_sample):
    if len(pop_sample) == 0:
        return 0
    return sum(p.similarity(new) for p in pop_sample) / len(pop_sample)

def is_diverse(avg_similarity, pressure):
    return avg_similarity < 1 - pressure

default_elite = 0
default_eq_ch = 0.5
default_max_sel_pressure = 200
default_div_threshold = 0.01

class Rapga(Gaos):
    name = "Rapga"

    def __init__(
        self,
        problem: Problem,
        solution_cls,
        popsize: int,
        mutation_rate: float,
        elite_size: int = default_elite,
        equal_chances: float = default_eq_ch,
        max_selective_pressure: int = default_max_sel_pressure,
        early_stop=True,
        diversity_threshold=default_div_threshold,
        min_popsize=None,
    ):
        super().__init__(
            problem,
            solution_cls,
            popsize,
            mutation_rate,
            elite_size,
            equal_chances,
            max_selective_pressure,
            early_stop,
        )
        self.diversity_threshold = diversity_threshold
        self.min_popsize = min_popsize or ( popsize // 10 + 2)

    def check_early_stop(self):
        return super().check_early_stop() or self.actual_popsize < self.min_popsize

    def solve(self, steps):
        self.problem.eval_count = 0
        popsize_sqrt = int(self.popsize ** (1/2)) + 1
        self._rank()
        while self.problem.eval_count < steps and self.actual_popsize:

            next_generation = self.population[: self.elite_size]

            while len(next_generation) < self.popsize:
                pressure = 0.8 * self.problem.eval_count / steps
                new = self._breed(popsize_sqrt, pressure=pressure)
                diversity_sample = (
                    next_generation
                    if len(next_generation) <= popsize_sqrt * 2
                    else random.sample(next_generation, popsize_sqrt * 2)
                )
                similarities = {}
                for n in new:
                    similarities[n] = average_similarity(n, diversity_sample)
                    n.regularization_score -= 0.2 * self.problem.score_span * similarities[n]

                next_generation += [
                    c
                    for c in new
                    if is_diverse(similarities[c], self.diversity_threshold)
                ]
                if self.selective_pressure == self.max_selective_pressure:
                    break

            self.population = next_generation

            self._rank()
            self.record()
            self.generation += 1
            if not self.generation % 5:
                print("Generation", self.generation, self.problem.eval_count)

            if self.early_stop and self.check_early_stop():
                break

        self.salut()

    def record(self):
        super().record()
        self.record_metric("actual popsize", self.actual_popsize)

    def __str__(self):
        string = f"{self.name} with mut_rate - {self.mutation_rate:.5f} & " \
            f"pop_size - {self.actual_popsize} "
        if self.elite_size != default_elite:
            string += f" & elite - {self.elite_size}"
        if self.equal_chances != default_eq_ch:
            string += f" & equal_c - {self.equal_chances}"
        if self.diversity_threshold != default_div_threshold:
            f" & div threshold - {self.diversity_threshold}"

        return string






from blackopt.algorithms import Rapga
from blackopt.abc import Problem, Solver
from typing import List, Tuple


import pathos


def iteration(inp: Tuple[Rapga, int]):
    ga, steps = inp
    ga.solution_cls.problem = ga.problem
    ga.generation = 0
    ga.record = lambda : None
    ga.problem.eval_count = 0
    ga.solve(steps)
    ga.metrics = None
    return ga


class Sasegasa(Solver):
    name = "Sasegasa"
    def __init__(
        self,
        problem: Problem,
        solution_cls,
        popsize: int,
        mutation_rate: float,
        elite_size: int = 0,
        equal_chances: float = 0.5,
        growth_factor=30,
        n_villages=12,
    ):
        self.problem = problem
        self.solution_cls = solution_cls
        self.popsize = popsize
        self.mutation_rate = mutation_rate
        self.elite_size = elite_size
        self.equal_chances = equal_chances
        self.growth_factor = growth_factor
        self.n_villages = n_villages

        super().__init__(problem, solution_cls)

        self.pool = pathos.pools.ProcessPool()

    def solve(self, steps):
        self.problem.eval_count = 0
        runs = sum(range(self.n_villages + 1))

        steps_per_run = steps // runs
        print(f"Sasegasa will run with {steps_per_run} steps per epoch/village")

        villages = [
            Rapga(
                self.problem,
                self.solution_cls,
                self.popsize,
                self.mutation_rate,
                self.elite_size,
                self.equal_chances,
                self.growth_factor,
                early_stop=True
            ) for i in range(self.n_villages)
        ]

        while len(villages) > 1:
            villages: List[Rapga] = self.pool.map(iteration, zip(villages, [steps_per_run] * len(villages)))
            self.problem.eval_count += sum(v.problem.eval_count for v in villages)
            self.best_solution = max(
                [s.best_solution for s in villages] + [self.best_solution], key=lambda x: x.score
            )
            self.record()
            print(f"Sasegasa: {len(villages)}, {self.problem.eval_count}")

            population = []
            for v in villages:
                population += v.population

            self.n_villages -= 1
            new_size = len(population) // self.n_villages
            populations = [population[i * new_size: i * new_size + new_size] for i in range(self.n_villages)]
            homeless = population[new_size * self.n_villages:]
            populations[-1] += homeless

            villages = villages[:-1]
            for v, p in zip(villages, populations):
                v.population = p
                v.popsize = len(p)

        v = villages[0]
        iteration( (v, steps_per_run*3) )
        self.best_solution = max(
            [v.best_solution] + [self.best_solution], key=lambda x: x.score
        )
        self.problem.eval_count += v.problem.eval_count
        self.record()






from blackopt.examples.problems.stepwise import StepSolution, StepProblem
from blackopt.examples.problems.bumpy import BumpyProblem, BumpySolution
from blackopt.examples.problems.tsp.tsp import TspProblem, TspSolution


import os
import abc
from typing import ClassVar, Dict, DefaultDict, SupportsFloat
from collections import defaultdict
import datetime

from ilya_ezplot import Metric
import dill

from blackopt.abc import Problem, Solution


class keydefaultdict(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


class Solver(abc.ABC):
    checkpoints_folder = "_checkpoints"

    name: str = None
    best_solution: Solution = None

    def __init__(self, problem: Problem, solution_cls: ClassVar[Solution]):
        problem.eval_count = 0
        self.problem = problem
        solution_cls.problem = problem
        self.solution_cls = solution_cls
        self.best_solution: Solution = self.solution_cls.random_solution()
        self.metrics: DefaultDict[str, Metric] = keydefaultdict(
            lambda k: Metric(name=str(self), y_label=k, x_label="evaluations")
        )

        self.checkpoint_path = os.path.join(self.checkpoints_folder, str(self.problem))


    def record(self):
        solution_metric_dict = self.best_solution.metrics()
        for k, v in solution_metric_dict.items():
            self.record_metric(f"best_{k}", v)

    def record_metric(self, name: str, val: SupportsFloat):
        self.metrics[name].add_record(self.problem.eval_count, val)

    @abc.abstractmethod
    def solve(self, steps):
        raise NotImplementedError()

    def checkpoint(self):
        os.makedirs(self.checkpoint_path, exist_ok=True)
        timestamp = datetime.datetime.now().strftime("%y-%m-%d_%H-%M-%S")

        with open(os.path.join(self.checkpoint_path, timestamp), 'wb') as f:
            dill.dump(self, f)

    @staticmethod
    def restore_latest(problem: Problem) -> 'Solver':
        directory = os.path.join(Solver.checkpoints_folder, str(problem))
        checkpoints = os.listdir(directory)

        if len(checkpoints) == 0:
            raise Exception(f"No checkpoints found in directory {directory}")
        else:
            cp = sorted(checkpoints)[-1]
            with open(os.path.join(directory, cp), 'rb') as f:
                restored = dill.load(f)
                assert isinstance(restored, Solver)
                return restored

    def __str__(self):
        return str(self.name)



import os
import datetime
from typing import List, Dict
from collections import defaultdict

from ilya_ezplot import Metric, plot_group

from blackopt.compare import SolverFactory
from blackopt.abc import Problem


def generate_report(
    problem: Problem, metrics: Dict[SolverFactory, Dict[str, Metric]], root_dir=None
):
    """ plot multiple curves from the metrics list"""
    timestamp = datetime.datetime.now().strftime("%m-%d_%H-%M-%S")
    if root_dir:
        problem_path = os.path.join(root_dir, "reports", str(problem))
    else:
        problem_path = os.path.join("reports", str(problem))

    m_groups = defaultdict(list)

    for sf, ms_dict in metrics.items():
        for key, m in ms_dict.items():
            m.discard_warmup(0.15)
            m_groups[key].append(m)

    for key, ms in m_groups.items():
        plot_group(ms, f"{problem_path}@{timestamp}", name=key, stdev_factor=0.1, smoothen=False)



from blackopt.abc.solver import Solver

class RandomSearch(Solver):
    name = "random search"

    def solve(self, steps):

        doc_freq = 1 + steps // 500

        for i in range(steps):
            solution = self.solution_cls.random_solution()
            if solution.score > self.best_solution.score:
                self.best_solution = solution

            if not i % doc_freq:
                print(i)
                self.record()


import pathos
import dill

class MulticoreRS(Solver):
    name = "multicore random search"

    def solve(self, steps):
        self.problem.eval_count = 0
        pool = pathos.pools.ProcessPool()
        n_cpus = pool.ncpus

        mapping_steps = int(steps ** (1 / 5))
        self.steps_per_pool = steps // (n_cpus * mapping_steps)

        print(self.steps_per_pool)
        print(mapping_steps)
        print(n_cpus)

        to_solve = dill.copy(self)

        for i in range(mapping_steps):
            solutions = pool.map(self._solve_pool, [to_solve] * n_cpus)
            self.best_solution = max(
                (s for s in solutions + [self.best_solution]), key=lambda x: x.score
            )
            self.problem.eval_count += self.steps_per_pool * n_cpus
            self.record()
            print(i * n_cpus * self.steps_per_pool)

    @staticmethod
    def _solve_pool(self: 'MulticoreRS'):

        doc_freq = 1 + self.steps_per_pool // 500

        for i in range(self.steps_per_pool):
            solution = self.solution_cls.random_solution()
            if solution.score > self.best_solution.score:
                self.best_solution = solution

            if not i % doc_freq:
                self.record()

        return self.best_solution



from typing import Callable, List, NoReturn
from blackopt.abc import Solver

from blackopt.log import get_logger


class ContinuousOptimizer:
    def __init__(
        self,
        problem,
        solver_factory: Callable[[], Solver],
        evals_per_step=1_000_000,
        step_callbacks: List[Callable[['ContinuousOptimizer'], NoReturn]]= None,
        max_repeats = 10
    ):
        self.problem = problem
        self.evals_per_step = evals_per_step
        self.logger = get_logger()
        self.steps = 0
        self.step_callbacks = step_callbacks or []
        self.max_repeats = max_repeats
        self.repeats = 0
        self.last_exception = "No exception so far"

        self.solver = solver_factory()

    def run(self):
        self.logger.info("starting optimization")
        while True:
            try:
                self.step()
                self.checkpoint()
            except Exception as e:
                if str(e) == self.last_exception:
                    self.repeats += 1
                    if self.repeats > self.max_repeats:
                        break
                else:
                    self.repeats = 0
                self.logger.warning(f"Encountered an exception: {e}")
                self.restore_latest()

    def step(self):
        for callback in self.step_callbacks:
            callback(self)

        self.solver.solve(self.evals_per_step)
        self.steps += 1
        self.logger.info(
            "step done", step=self.steps, score=self.solver.best_solution.score
        )

    def checkpoint(self):
        self.solver.checkpoint()

    def restore_latest(self):
        self.solver = Solver.restore_latest(self.problem)



from blackopt.algorithms.genetic_algorithm import GeneticAlgorithm
from blackopt.algorithms.genetic_algorithm_injection import GeneticAlgorithmInject
from blackopt.algorithms.genetic_algo_os import Gaos
from blackopt.algorithms.genetic_algo_rapga import Rapga
from blackopt.algorithms.sasegasa import Sasegasa
from blackopt.algorithms.sasegasa_continuous import SasegasaContinuous

from blackopt.algorithms.multicore_genetic_algorithm import MulticoreGeneticAlgorithm
from blackopt.algorithms.hill_climb import HillClimber
from blackopt.algorithms.random_search import RandomSearch, MulticoreRS
from blackopt.algorithms.simulated_annealing import SimAnneal


from blackopt.abc import Solver, Problem
from typing import ClassVar


class HillClimber(Solver):
    name = "hill climb"

    def __init__(self, problem: Problem, solution_cls: ClassVar, mutation_rate):

        assert 0 < mutation_rate <= 1
        self.mutation_rate = mutation_rate
        super().__init__(problem, solution_cls)


    def solve(self, steps):
        self.problem.eval_count = 0
        doc_freq = 1 + steps // 500

        for i in range(steps):
            solution = self.best_solution.mutate(self.mutation_rate)
            if solution.score > self.best_solution.score:
                self.best_solution = solution

            if i % doc_freq == 0:
                self.record()


    def __str__(self):
        return f"{self.name} {self.mutation_rate}"



from blackopt.examples.problems import TspProblem, TspSolution
from blackopt.algorithms import RandomSearch
from blackopt.algorithms import GeneticAlgorithm
from blackopt.util.document import generate_report


from blackopt.compare import compare_solvers, SolverFactory

n_steps = int(2e3)
n_trials = 3

problems = [
    TspProblem.random_problem(n_dim, cities)
    for n_dim in [2, 10, 50]
    for cities in [40, 200, 1000]
]

for problem in problems:

    solvers = []
    solvers.append(SolverFactory(RandomSearch, problem, TspSolution))
    for popsize in [10, 50]:
        for elite_size in [1, popsize // 10]:
            for mr in [5e-4, 5e-3, 0.01]:
                # if heavy:
                #     style["dashes"] = [2, 2, 10, 2]
                # style["linewidth"] = linewidth
                # style["color"] = color
                ga = SolverFactory(
                    GeneticAlgorithm,
                    problem,
                    TspSolution,
                    popsize,
                    mr,
                    elite_size,
                )
                solvers.append(ga)

    ms = compare_solvers(n_trials, n_steps, solvers)
    generate_report(problem, ms)



import abc
import os
import dill
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from blackopt.abc import Solution

import uuid


class Problem(abc.ABC):
    eval_count: int = None
    score_span: float = 1

    store_dir = "_problems"

    @abc.abstractmethod
    def evaluate(self, s: 'Solution') -> float:
        raise NotImplementedError()

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError()

    def save(self):
        os.makedirs(self.store_dir, exist_ok=True)
        identifier = str(self)

        existing = os.listdir(self.store_dir)
        if identifier in existing:
            identifier += uuid.uuid4()

        with open(os.path.join(self.store_dir, identifier), 'wb') as f:
            dill.dump(self, f)
            print(f"Stored problem as {identifier}")

    @staticmethod
    def load(identifier: str) -> 'Problem':
        directory = os.path.join(Problem.store_dir)

        with open(os.path.join(directory, identifier), 'rb') as f:
            restored = dill.load(f)
            assert isinstance(restored, Problem)
            return restored





from blackopt.examples.problems import TspProblem, TspSolution
from blackopt.algorithms import (
    RandomSearch,
    MulticoreRS,
    MulticoreGeneticAlgorithm,
    Sasegasa,
    Rapga,
    SasegasaContinuous,
)
from blackopt.util.document import generate_report


import time

# problem = TspProblem.random_problem(2, 130)
# problem.save()
problem = TspProblem.load("Tsp 130 cities & 2 dim")


solver = Rapga(problem, TspSolution, 100, 1 / 130)
t = time.time()
solver.solve(int(2e5))
print("Duration: ", time.time() - t)
generate_report(problem, {solver: solver.metrics})

# solver = SasegasaContinuous(problem, TspSolution, 200, 1 / 130)
# for i in range(10):
#     t = time.time()
#     solver.solve(int(2e5))
#     print(f"Step {i}, Duration: ", time.time() - t)
#
# generate_report(problem, {solver: solver.metrics})



from blackopt.examples.problems import TspProblem, TspSolution
from blackopt.algorithms import SasegasaContinuous

from blackopt.persistence import ContinuousOptimizer
from blackopt.compare import SolverFactory


# problem = TspProblem.random_problem(2, 130)
problem = TspProblem.load("Tsp 130 cities & 2 dim")

solver = SolverFactory(SasegasaContinuous, problem, TspSolution, 200, 1/130, 0)

co = ContinuousOptimizer(problem, solver)
co.run()



from blackopt.algorithms import Rapga
from blackopt.abc import Problem, Solver
from typing import List, Tuple

import random

import pathos


def iteration(inp: Tuple[Rapga, int]):
    ga, steps = inp
    ga.solution_cls.problem = ga.problem
    ga.generation = 0
    ga.record = lambda: None
    ga.problem.eval_count = 0
    ga.solve(steps)
    ga.metrics = None
    return ga


class SasegasaContinuous(Solver):
    name = "SasegasaContinuous"

    def __init__(
        self,
        problem: Problem,
        solution_cls,
        popsize: int,
        mutation_rate: float,
        elite_size: int = 0,
        equal_chances: float = 0.5,
        max_selective_pressure: int = 200,
        diversity_threshold = 0.01,
        n_villages = None


    ):
        self.problem = problem
        self.solution_cls = solution_cls
        self.popsize = popsize
        self.mutation_rate = mutation_rate
        self.elite_size = elite_size
        self.equal_chances = equal_chances
        self.max_selective_pressure = max_selective_pressure
        self.pool = pathos.pools.ProcessPool()
        self.n_villages = n_villages or int(self.pool.ncpus * 1.2)

        self.villages = [
            Rapga(
                self.problem,
                self.solution_cls,
                self.popsize,
                self.mutation_rate,
                self.elite_size,
                self.equal_chances,
                diversity_threshold=diversity_threshold,
                max_selective_pressure = max_selective_pressure,
                early_stop=True,
            )
            for i in range(self.n_villages)
        ]

        super().__init__(problem, solution_cls)


    @property
    def population(self):
        population = []
        for v in self.villages:
            random.shuffle(v.population)
            population += v.population
        return population


    def solve(self, steps):

        steps_per_run = steps // self.n_villages
        print(f"Sasegasa will run with {steps_per_run} steps per epoch/village")


        self.villages: List[Rapga] = self.pool.map(
            iteration, zip(self.villages, [steps_per_run] * len(self.villages))
        )
        self.best_solution = max(
            [s.best_solution for s in self.villages] + [self.best_solution],
            key=lambda x: x.score,
        )

        self.problem.eval_count += sum(v.problem.eval_count for v in self.villages)
        self.record()

        population = self.population
        shift = random.randint(1, len(population))
        population = population[shift:] + population[:shift]
        populations = [
            population[i * self.popsize : i * self.popsize + self.popsize]
            for i in range(self.n_villages)
        ]
        homeless = population[self.popsize * self.n_villages :]
        populations[-1] += homeless

        for v, p in zip(self.villages, populations):
            v.population = p



import numpy as np
from functools import lru_cache

np.random.seed(1)


class City:

    n_cities = 0
    city_by_id = {}

    def __init__(self, ndim):
        self.uid = City.n_cities
        City.city_by_id[self.uid] = self
        City.n_cities += 1
        self.coordinates = np.random.random(size=[ndim])

    @lru_cache(maxsize=int(2**20))
    def distance(self, city: 'City'):
        return np.linalg.norm(self.coordinates - city.coordinates)

    def __eq__(self, other):
        return self.uid == other.uid

    def __hash__(self):
        return self.uid

    def __repr__(self):
        return f"City {self.uid}"



from blackopt.examples.problems import TspProblem, TspSolution
from blackopt.algorithms import RandomSearch, HillClimber, Gaos, Rapga, Sasegasa
from blackopt.algorithms import GeneticAlgorithm, SimAnneal
from blackopt.util.document import generate_report


from blackopt.compare import compare_solvers, SolverFactory

n_steps = int(4e5)
n_trials = 6

cities = 60
problem = TspProblem.random_problem(2, cities)

solvers = []

solvers.append(
    SolverFactory(
        Rapga, problem, TspSolution, 300, 2 / cities, 0, 0.5
    )
)
solvers.append(
    SolverFactory(
        Rapga, problem, TspSolution, 300, 2 / cities, 0, 0.5, regularize=True
    )
)


if __name__ == "__main__":
    import time

    t = time.time()
    ms = compare_solvers(n_trials, n_steps, solvers)
    generate_report(problem, ms)
    print(time.time() - t)



import datetime, logging, sys
import os
from structlog import wrap_logger
from structlog.processors import JSONRenderer, TimeStamper

from logging.handlers import RotatingFileHandler


logging.basicConfig(stream=sys.stdout, format="%(message)s")


BACKUP_COUNT = 5
MAX_FILE_SIZE = 20000 * 1000  # bytes
LOGDIR = "_logs"



def get_logger(name="blackopt", logdir=LOGDIR, **initial_values):

    path = os.path.join(logdir, name)

    os.makedirs(os.path.dirname(path), exist_ok=True)

    def rotating_file_handler(log_path):
        return RotatingFileHandler(log_path, mode='a',
                                    maxBytes=BACKUP_COUNT * MAX_FILE_SIZE,
                                    backupCount=BACKUP_COUNT)

    file_hdlr = rotating_file_handler(path)
    if os.path.exists(path):
        try:
            if os.stat(path).st_size != 0:
                file_hdlr.doRollover()
        except FileNotFoundError:
            # Rollover failed. prefer to overwrite old log instead terminating Moonfish.
            file_hdlr = rotating_file_handler(path)


    logger = logging.getLogger(name)
    logger.addHandler(file_hdlr)
    logger.setLevel(logging.INFO)

    log = wrap_logger(
        logger,
        processors=[
            # filter_by_level,
            TimeStamper(fmt="iso"),
            JSONRenderer(sort_keys=True),
        ],
        **initial_values
    )



    return log

if __name__ == "__main__":
    log = get_logger()
    log.info("something.filtered")
    log.warning("something.not_filtered")



from setuptools import setup

version = 0.19


def get_requirements():
    with open("requirements.txt") as fp:
        return fp.read()


setup(
    name="blackopt",
    packages=[
        "blackopt",
        "blackopt.abc",
        "blackopt.algorithms",
        "blackopt.examples",
        "blackopt.examples.launchers",
        "blackopt.examples.problems",
        "blackopt.examples.problems.tsp",
        "blackopt.util",
    ],
    # package_dir={"": "."},
    version=version,
    description="black box optimization",
    author="Ilya Kamenshchikov",
    keywords=["optimization"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    license="MIT",
    long_description="In a far, far away galaxy...",
    install_requires=get_requirements(),
    python_requires=">=3.6",
    url="https://github.com/ikamensh/optim",
)



from collections import defaultdict
from typing import List

import numpy as np

from blackopt.abc import Problem, Solution
from blackopt.abc.solver import Solver
from blackopt.algorithms import GeneticAlgorithm


class GeneticAlgorithmInject(GeneticAlgorithm):
    name = "GA"

    def __init__(
        self,
        problem: Problem,
        solution_cls,
        popsize: int,
        mutation_rate: float,
        elite_size: int,
        inject_quality: int,
    ):
        super().__init__(problem, solution_cls, popsize, mutation_rate, elite_size)
        self.inject_quality = inject_quality

    def _inject(self):
        best_solution = self.solution_cls.random_solution()
        for i in range(self.inject_quality):
            solution = self.solution_cls.random_solution()
            if solution.score > best_solution.score:
                best_solution = solution

        return best_solution

    def solve(self, steps):
        self.problem.eval_count = 0
        while self.problem.eval_count < steps:

            next_generation = self.population[: self.elite_size]
            next_generation += [self._inject()]
            next_generation += self._breed(self.popsize - self.elite_size - 1)
            self.population = next_generation

            self._rank()
            self.record()
            self.generation += 1
            if not self.generation % 100:
                print("Generation", self.generation, self.problem.eval_count)

        print(f"{self} is Done in {self.generation} generations")

    def _breed(self, n: int, smoothen_chances=0) -> List[Solution]:

        parents = self._select_parents(n, smoothen_chances)
        children: List[Solution] = []

        for i in range(n):
            parent_1 = parents[i]
            parent_2 = parents[len(parents) - i - 1]
            children += parent_1.crossover(parent_2)

        children = [child.mutate(self.mutation_rate) for child in children]

        return children

    def __str__(self):
        return (
            f"{self.name} with mut - {self.mutation_rate} & "
            f"pop - {self.popsize} & "
            f"elite - {self.elite_size}"
            f"inj - {self.inject_quality}"
        )



from typing import List
import random
from functools import lru_cache

from blackopt.abc import Problem, Solution


class StepProblem(Problem):
    def __init__(self, thresholds: List[float]):
        self.n_dim = len(thresholds)
        self.score_span = len(thresholds)
        self.thresholds = thresholds
        for t in thresholds:
            assert 0 <= t <= 1
        self.eval_count = 0

    @staticmethod
    def random_problem(n_dim: int):
        thresholds = [random.random() for i in range(n_dim)]
        return StepProblem(thresholds)

    @staticmethod
    def _step_function(values: List[float], thresholds: List[float]) -> int:
        result = 0

        for val, threshold in zip(values, thresholds):
            if val > threshold:
                result += 1

        return result

    def evaluate(self, s: 'StepSolution') -> int:
        self.eval_count += 1
        return self._step_function(s.genes, self.thresholds)

    def __str__(self):
        return f"{self.__class__.__name__} {self.n_dim}"


class StepSolution(Solution):
    def __init__(self, values):
        self.genes = values

    @staticmethod
    def random_solution() -> Solution:
        values = [random.random() for i in range(StepSolution.problem.n_dim)]
        return StepSolution(values)

    def mutate(self, rate: float):
        new_values = []
        for v in self.genes:
            if random.random() < rate:
                new_values.append(random.random())
            else:
                new_values.append(v)

        return StepSolution(new_values)

    def crossover(self, other: 'StepSolution'):
        crossover_point = random.randint(1, len(self.genes) - 1)

        child_a = self.genes[:crossover_point] + other.genes[crossover_point:]
        child_b = other.genes[:crossover_point] + self.genes[crossover_point:]

        return [StepSolution(child_a), StepSolution(child_b)]

    @lru_cache(maxsize=512)
    def similarity(self, other: 'StepSolution'):

        diff = 0
        for i in range(self.problem.n_dim):
            diff += abs(self.genes[i] - other.genes[i])

        return 1 - diff / self.problem.n_dim






from typing import List
import random
import math
from functools import lru_cache

from blackopt.abc import Problem, Solution
from blackopt.examples.problems.tsp.city import City



class TspProblem(Problem):
    name = "Tsp"

    def __init__(self, cities: List[City]):
        self.cities = cities
        self.n_dim = len(cities[0].coordinates)
        self.eval_count = 0
        self.score_span = self.max_dist = len(self.cities) * math.sqrt(self.n_dim)

    @staticmethod
    def random_problem(n_dim: int, cities: int):
        cities = [City(n_dim) for _ in range(cities)]
        return TspProblem(cities)

    def evaluate(self, s: 'TspSolution'):
        self.eval_count += 1
        return self.max_dist - self.route_distance(s.route)

    @staticmethod
    def route_distance(route: List[City]):
        pathDistance = 0
        path = route + [route[0]]
        for i in range(0, len(path) - 1):
            pathDistance += path[i].distance(path[i + 1])
        return pathDistance

    def __str__(self):
        return f"Tsp {len(self.cities)} cities & {self.n_dim} dim"


class TspSolution(Solution):
    def __init__(self, route: List[City]):
        self.route = route
        assert set(route) == set(self.problem.cities)

    @staticmethod
    def random_solution() -> 'TspSolution':
        cpy = list(TspSolution.problem.cities)
        random.shuffle(cpy)
        return TspSolution(cpy)

    def mutate(self, mutationRate: float) -> 'TspSolution':
        route = list(self.route)

        for i in range(len(route)):
            if random.random() < mutationRate:
                j = int(random.random() * len(route))
                route[i], route[j] = route[j], route[i]

        return TspSolution(route)

    def crossover(self, other: 'TspSolution'):
        crossover_point = random.randint(1, len(self.route) - 1)

        child_left = self.route[:crossover_point]
        check = {city.uid for city in child_left}
        child_right = [city for city in other.route if city.uid not in check]

        return [TspSolution(child_left + child_right)]

    @lru_cache(maxsize=512)
    def similarity(self, other: 'TspSolution'):
        r1 = self.route
        r2 = other.route

        moves_1 = { (r1[i], r1[i+1]) for i in range(len(r1)-1)}
        moves_1.add( (r1[-1], r1[0]) )
        moves_2 = { (r2[i], r2[i+1]) for i in range(len(r2)-1)}
        moves_2.add((r2[-1], r2[0]))

        return len( moves_1.intersection(moves_2) ) / len(moves_1)
